import cv2
import numpy as np

def detect_charuco_diamond(gray, aruco_dict, square_length, marker_length, target_diamond_ids, *, verbose=0):
    # detect Aruco
    detected_corners, detected_ids, rejectedImgPoints = cv2.aruco.detectMarkers(gray, aruco_dict)
    if detected_ids is None:
        if verbose>0: print('Aruco markers are not found')
        return None, None
    if len(detected_ids) < 4:
        if verbose>0: print(f'Four Aruco markers are not found. Detected IDs are: {detected_ids.flatten()}')
        return None, None
    
    # detect Aruco diamonds
    detected_diamond_corners, detected_diamond_ids = cv2.aruco.detectCharucoDiamond(gray, detected_corners, detected_ids, square_length / marker_length)
    if detected_diamond_ids is None:
        if verbose>0: print('Diamond markers are not found')
        return None, None

    # check if the detected one is the target
    target_diamond_ids = np.array(target_diamond_ids).reshape((-1, 1, 4)) # detectCharucoDiamond returns (N, 1, 4) array
    ret_corners = []
    ret_ids = []
    for tdi in target_diamond_ids:
        found = False
        for ddc, ddi in zip(detected_diamond_corners, detected_diamond_ids):
            if np.array_equal(tdi, ddi):
                ret_corners.append(ddc)
                ret_ids.append(ddi)
                found = True
                break
        if not found:
            if verbose>0: print(f'Diamond {tdi.flatten()} not found')
            return None, None

    return np.array(ret_corners), np.array(ret_ids)





import unittest
import numpy as np
import cv2
import pycalib

class TestPyCalibMirror(unittest.TestCase):
    def prepare(self):
        DATA_DIR = './data/mirror/chess/'
        model = np.loadtxt(f'{DATA_DIR}/model.txt')
        #print(f'3D points (#_of_reference_points x 3): {model.shape}')

        input1 = np.loadtxt(f'{DATA_DIR}/input1.txt')
        input2 = np.loadtxt(f'{DATA_DIR}/input2.txt')
        input3 = np.loadtxt(f'{DATA_DIR}/input3.txt')
        input4 = np.loadtxt(f'{DATA_DIR}/input4.txt')
        input5 = np.loadtxt(f'{DATA_DIR}/input5.txt')
        input = np.array([input1, input2, input3, input4, input5])
        #print(f'2D points (#_of_mirrors x #_of_reference_points x 2)): {input.shape}')

        K = np.loadtxt(f'{DATA_DIR}/camera.txt')
        #print(f'K = {K}')

        return model, input, K

    def test_mirror_5(self):
        model, input, K = self.prepare()

        # linear solution
        R0, T0, n0, d0, rep0 = pycalib.mirror.tnm(model, input, K)
        #print(f'Reprojection error (linear) = {rep0:.3} px')
        self.assertLess(rep0, 6.284716397404586)

        # non-linear refinement
        R, T, n, d, rep = pycalib.mirror.tnm_ba(model, input, K, R0, T0, n0, d0)
        #print(f'Reprojection error (BA) = {rep:.3} px')
        self.assertLess(rep, 0.6401348975172396)

    def test_mirror_3(self):
        model, input, K = self.prepare()
        input = input[:3]

        # linear solution
        R0, T0, n0, d0, rep0 = pycalib.mirror.tnm(model, input, K)
        #print(f'Reprojection error (linear) = {rep0:.3} px')
        self.assertLess(rep0, 1.51)

        # non-linear refinement
        R, T, n, d, rep = pycalib.mirror.tnm_ba(model, input, K, R0, T0, n0, d0)
        #print(f'Reprojection error (BA) = {rep:.3} px')
        self.assertLess(rep, 0.689)

    def test_mirror_mask(self):
        model, input, K = self.prepare()

        TH = 25
        rng = np.random.default_rng(0)
        for x in input:
            sign = rng.integers(0, 100, len(x))
            x[sign < TH, :] = np.nan
        # print(x)

        # print(f'{TH}% of observations are randomly masked')

        # linear solution
        R0, T0, n0, d0, rep0 = pycalib.mirror.tnm(model, input, K)
        #print(f'Reprojection error (linear) = {rep0:.3} px')
        self.assertLess(rep0, 6.493891993463244)

        # non-linear refinement
        R, T, n, d, rep = pycalib.mirror.tnm_ba(model, input, K, R0, T0, n0, d0)
        #print(f'Reprojection error (BA) = {rep:.3} px')
        self.assertLess(rep, 0.6538855436697778)



__all__ = ["ba", "bal",  "calib", "plot", "util", "sphere", "diamond", "timestamp", "mirror", "circle", "robust"]
__version__ = '2025.4.19.1'

from .ba import *
from .bal import *
from .calib import *
from .plot import *
from .util import *
from .sphere import *
from .diamond import *
from .timestamp import *
from .mirror import tnm, tnm_ba, householder
from .circle import *
from .robust import *




import numpy as np
import matplotlib.pyplot as plt
import cv2

def axisEqual3D(ax):
    """
    https://stackoverflow.com/a/19248731
    """
    extents = np.array([getattr(ax, 'get_{}lim'.format(dim))() for dim in 'xyz'])
    sz = extents[:,1] - extents[:,0]
    centers = np.mean(extents, axis=1)
    maxsize = max(abs(sz))
    r = maxsize/2
    for ctr, dim in zip(centers, 'xyz'):
        getattr(ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)
        getattr(ax, 'set_{}label'.format(dim))(dim)

def plotPoly(ax, corners_Nx3, **kwargs):
    """Plot a polygon in 3D

    Args:
        ax (Axis3D): target axis
        corners_Nx3: Nx3 corner points
    """
    corners = np.vstack((corners_Nx3, corners_Nx3[0]))
    ax.plot(corners[:,0], corners[:,1], corners[:,2], **kwargs)

def plotCamera(ax, R, t, *, color=None, scale=1, width=2, height=1.5, focal_length=3, label=None):
    """Plot a camera in 3D

    Args:
        ax (Axis3D): target axis
        R: 3x3 c2w rotation matrix
        t: 3x1 c2w translation vector
        c: color string
        scale: scaling factor
        width: horizontal size of the camera screen
        height: vertical size of the camera screen
        focal_length: height of the camera cone
    """
    t = t.reshape((3, 1))

    w = width
    h = height
    f = focal_length

    # focus, br, tr, tl, bl, triangle
    ps_c = np.array(([0,0,0], [w/2,h/2,f], [w/2,-h/2,f], [-w/2,-h/2,f], [-w/2,h/2,f], [0,-h/2-h/4,f]))
    ps_w = (scale * R @ ps_c.T + t).T

    L01 = np.array([ps_w[0], ps_w[1]])
    L02 = np.array([ps_w[0], ps_w[2]])
    L03 = np.array([ps_w[0], ps_w[3]])
    L04 = np.array([ps_w[0], ps_w[4]])
    L1234 = np.array([ps_w[1], ps_w[2], ps_w[3], ps_w[4], ps_w[1]])
    L253 = np.array([ps_w[2], ps_w[5], ps_w[3]])

    p = ax.plot(L01[:,0], L01[:,1], L01[:,2], "-", color=color)
    if color is None:
        color = p[-1].get_color()
    ax.plot(L02[:,0], L02[:,1], L02[:,2], "-", color=color)
    ax.plot(L03[:,0], L03[:,1], L03[:,2], "-", color=color)
    ax.plot(L04[:,0], L04[:,1], L04[:,2], "-", color=color)
    ax.plot(L1234[:,0], L1234[:,1], L1234[:,2], "-", color=color)
    ax.plot(L253[:,0], L253[:,1], L253[:,2], "-", color=color)

    if label is not None:
        ax.text(ps_w[0][0], ps_w[0][1], ps_w[0][2], label)


def plotCameras(camera_params, points_3d, *, scale=-1, title=None, draw_zplane=False, label=None):
    """Plot cameras and points in 3D

    Args:
        camera_params: Nx17 camera parameters (rvec, tvec, f, ...)
        points_3d: Mx3 3D points
    """
    assert points_3d.shape[1] == 3
    Nc = len(camera_params)

    R = []
    t = []
    p = []
    for c in camera_params:
        R.append(cv2.Rodrigues(c[:3])[0])
        t.append(c[3:6])
        p.append(- R[-1].T @ t[-1][:,None])
    p = np.array(p)

    if scale <= 0:
        scale = 0.05
        if Nc != 1:
            l = np.linalg.norm(t[0] - t[1])
            scale *= l

    fig = plt.figure()
    ax = plt.axes(projection='3d')
    fig.add_axes(ax)
    ax.set_xlim(p[:,0].min(), p[:,0].max())
    ax.set_ylim(p[:,1].min(), p[:,1].max())
    ax.set_zlim(p[:,2].min(), p[:,2].max())
    #ax.set_ylim(-lim, lim)
    #ax.set_zlim(0, lim)
    ax.plot(points_3d[:,0], points_3d[:,1], points_3d[:,2], "o")
    cmap = plt.get_cmap("tab20")
    if label is None:
        label = [ f'CAM{i}' for i in range(Nc) ]
    for i in range(Nc):
        plotCamera(ax, R[i].T, p[i], color=cmap(i), scale=scale, label=label[i])
        #plotCamera(ax, R[i].T, - R[i].T @ t[i][:,None], cmap(i), scale)
    #plt.savefig('a.png')

    if title is not None:
        ax.text2D(0.05, 0.95, title, transform=ax.transAxes)
    if draw_zplane:
        s = np.linalg.norm(t[0] - t[1])
        ax.plot_surface(*np.meshgrid([-s, s], [-s, s]), np.zeros((2,2)), alpha=0.2)
        ax.plot([0, s], [0, 0], [0, 0])
        ax.plot([0, 0], [0, s], [0, 0])
        ax.text(0, 0, 0, 'O')
        ax.text(s, 0, 0, 'X')
        ax.text(0, s, 0, 'Y')

    axisEqual3D(ax)
    #fig.show()
    return fig

def plotMirror(ax, objpts, n, d, label):
    assert objpts.shape[1] == 3
    t = - objpts @ n.reshape((3,1)) - d
    p = objpts - t.reshape((-1,1)) * n.reshape((1,3))
    ax.plot(p[:,0], p[:,1], p[:,2], label=label)



import unittest
import os
import numpy as np
import pycalib
import time
import cv2

class TestPyCalibSphere(unittest.TestCase):

    def test_sphere(self):
        # dummy data
        W = 640
        H = 480
        fx = 320
        fy = fx
        u0 = W // 2
        v0 = H // 2
        K = np.array( [[ fx, 0, u0 ], [0, fy, v0], [0, 0, 1]])
        RADIUS = 0.5

        X_gt_all = []
        c_gt_all = []
        img_all = []
        cont_all = []
        for X in range(-8, 9, 2):
            for Y in range(-5, 6, 2):
                X_gt = np.array([X, Y, 10])
                c_gt = K @ X_gt.reshape((3,1))
                c_gt = (c_gt / c_gt[2]).flatten()
        
                img = pycalib.render_sphere(X_gt, RADIUS, K, W, H)
                contours, _ = cv2.findContours(img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        
                X_gt_all.append(X_gt)
                c_gt_all.append(c_gt)
                img_all.append(img)
                cont_all.append(contours[0])
        
        for X_gt, c_gt, img, cnt in zip(X_gt_all, c_gt_all, img_all, cont_all):
            ellipse = cv2.fitEllipse(cnt)
            c = pycalib.calc_sphere_center_from_ellipse(ellipse, K)
            self.assertLess(np.linalg.norm(c[:2] - c_gt[:2]), 1.0)

            c3d = pycalib.fit_sphere_center_3d_to_ellipse(cnt, K)
            c = K @ c3d.T
            c = c / c[2]
            self.assertLess(np.linalg.norm(c[:2] - c_gt[:2]), 1.0)

            c3d = pycalib.fit_sphere_center_3d_to_ellipse(cnt, K, resample=True)
            c = K @ c3d.T
            c = c / c[2]
            self.assertLess(np.linalg.norm(c[:2] - c_gt[:2]), 1.0)




# This is based on the Python version of the original implementation
#
#   https://github.com/computer-vision/takahashi2012cvpr
#

import numpy as np
import cv2
import scipy as sp
import itertools

def w2c(Xp: np.ndarray, rvec: np.ndarray, tvec:np.ndarray) -> np.ndarray:
    R = cv2.Rodrigues(rvec)[0]
    x = R @ Xp.T + tvec.reshape((3, 1))
    return x.T

def sub_tmn_orth(Cp_candidates: np.ndarray) -> np.ndarray:
    num_mirrors = len(Cp_candidates)
    assert Cp_candidates.shape[2:] == (3, 3), Cp_candidates.shape
    best_rho = np.inf
    best_cp = None
    for cp in itertools.product(*Cp_candidates):
        rho = 0
        for cp1, cp2 in itertools.combinations(cp, 2):
            Q = cp1 - cp2
            M = Q.T @ Q
            D, _ = np.linalg.eig(M)
            rho += D.min() / np.sum(D)
        if rho < best_rho:
            best_rho = rho
            best_cp = cp
    return best_cp

def make_A(Xp, n):
    num_mirrors = len(n)
    num_points = len(Xp)
    assert n.shape == (num_mirrors, 3), n.shape
    assert Xp.shape == (num_points, 3), Xp.shape

    A = []
    for i in range(num_mirrors):
        A_normal = np.zeros((3,num_mirrors))
        A_normal[:,i] = 2 * n[i]
        for j in range(num_points):
            A_sub = np.hstack([np.eye(3), A_normal, Xp[j, 0]*np.eye(3), Xp[j,1]*np.eye(3)])
            A.append(A_sub)
    A = np.vstack(A)
    return A

def make_B(Cp, n):
    num_mirrors = len(n)
    num_points = Cp.shape[1]
    assert n.shape == (num_mirrors, 3), n.shape
    assert Cp.shape == (num_mirrors, num_points, 3), Cp.shape

    B = []
    for i in range(num_mirrors):
        for j in range(num_points):

            ni = n[i].reshape((3,1))
            cij = Cp[i,j].reshape((3,1))

            b = -2 * (ni.T @ cij) * ni + cij
            B.append(b.flatten())
    B = np.concatenate(B)
    return B

def sub_tnm_rt(Xp: np.ndarray, Cp: np.ndarray):
    num_mirrors = len(Cp)
    # compute the axis vector m for each pair of mirror pose.
    m_all = [[] for _ in range(num_mirrors)]
    for i1 in range(num_mirrors-1):
        for i2 in range(i1+1, num_mirrors):
            Cp_1 = Cp[i1]
            Cp_2 = Cp[i2]
            # compute the Q matrix
            Q = Cp_2 - Cp_1
            # compute the M matrix
            M = Q.T @ Q
            # compute the eigen vector of M
            D, V = np.linalg.eig(M)
            # compute the m vector as a eigen vector for smallest eigen value.
            m = V[:, D.argmin()]

            m_all[i1].append(m)
            m_all[i2].append(m)

    n_all = []
    for m in m_all:
        # compute the S matrix
        S = np.array(m) # M-1 x 3
        # compute the eigen value of S
        D, V = np.linalg.eig(S.T @ S)
        # compute the normal vector as a eigen vector for smallest eigen value.
        n = V[:, D.argmin()]
        # solve the sign ambiguity
        if n[2] > 0:
            n = - n
        n_all.append(n)
    n_all = np.array(n_all)

    A = make_A(Xp, n_all)
    B = make_B(Cp, n_all)
    X = np.linalg.solve(A.T@A, A.T@B)

    T = X[:3]
    d = X[3:3+num_mirrors]
    r1 = X[3+num_mirrors:6+num_mirrors]
    r2 = X[6+num_mirrors:]

    r1 = r1 / np.linalg.norm(r1)
    r2 = r2 / np.linalg.norm(r2)
    r3 = np.cross(r1, r2)
    r3 = r3 / np.linalg.norm(r3)
    Q = np.array([r1, r2, r3]).T
    U, S, Vh = np.linalg.svd(Q)
    R = U @ Vh

    #print('R', R)
    #print('T', T)
    #print('d', d)
    return R, T, n_all, d

def householder(n, d):
    #  H = [ eye(3) - 2 * n * n', -2 * d * n; zeros(1,3), 1];
    n = n.reshape((3, 1))
    H = np.eye(4)
    H[:3, :3] = np.eye(3) - 2 * (n @ n.T)
    H[:3,  3] = (-2 * d * n).flatten()
    return H

def sub_reproj_single(Xp, q, R, T, n, d, K):
    N = len(Xp)
    assert Xp.shape == (N, 3), Xp.shape
    assert q.shape == (N, 2), q.shape
    assert R.shape == (3, 3), R.shape
    assert K.shape == (3, 3), K.shape

    # 4x4 householder transformation matrix
    H = householder(n, d)

    # reference points computed with estimated parameters.
    temp_Cp_h = np.ones((4, len(Xp)))
    temp_Cp_h[:3, :] = R @ Xp.T + T.reshape((3, 1))
    Cp_h = H @ temp_Cp_h

    # project the reference points to the image plane
    temp_q = K @ Cp_h[:3,:]
    temp_q = temp_q[:2,:] / temp_q[2,:]

    # mask errors by NaN at negative 2D observations (we should not use 0 here, as we cannot distingish if the error is actually zero or not, and also fakes the average error to be smaller as the mask grows.)
    diff = q - temp_q[:2].T
    diff[np.isnan(q[:,0]),:] = np.nan
    return diff

def sub_reproj_vec(Xp, q, R, T, n, d, K):
    num_mirrors = len(n)
    num_points = len(Xp)

    assert R.shape == (3, 3), R.shape
    assert T.shape == (3,) or T.shape == (3, 1), T.shape
    assert n.shape == (num_mirrors, 3) or n.shape == (num_mirrors, 3, 1), n.shape
    assert len(d) == num_mirrors
    assert q.shape == (num_mirrors, num_points, 2), q.shape
    assert K.shape == (3, 3)
    assert Xp.shape == (num_points, 3), Xp.shape

    err = []
    for i in range(num_mirrors):
        rep = sub_reproj_single(Xp, q[i], R, T, n[i], d[i], K)
        err.extend(rep)

    return np.array(err)

def sub_reproj(Xp, q, R, T, n, d, K):
    err = sub_reproj_vec(Xp, q, R, T, n, d, K)
    assert err.shape[1] == 2
    err = np.linalg.norm(err, axis=1)
    assert len(err) == len(Xp) * len(q)
    return np.nanmean(err)   # err may have NaN for masked observations

def tnm(Xp: np.ndarray, q: np.ndarray, K: np.ndarray, *, flags=None) -> [np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Finds the camera and the mirror poses from N>=3 input images

    Parameters
    ----------
    Xp : numpy.ndarray
        Nx3 matrix representing 3D coordinates of N reference points in the reference coordinate system.
    q : numpy.ndarray
        MxNx2 matrix representing 2D projections of N reference points obeserved via M mirrors.
        NaN values indicate "masked / invalid" points.
    K : numpy.ndarray
        3x3 intrinsic camera parameter.
    flags : int
        flags for cv2.solvePnP

    Returns
    -------
    R : numpy.ndarray
        3x3 rotation matrix of the camera.
    T : numpy.ndarray
        3x1 translation vector of the camera.
    n : numpy.ndarray
        Mx3 normal vectors of the mirror planes.
    d : numpy.ndarray
        Mx1 distances between the camera and the mirrors.
    rep : float
        The average of the reprojection errors.
    """

    # Num of mirrors
    M = q.shape[0]
    assert M >= 3, M
    
    # Num of reference points
    N = Xp.shape[0]
    assert N >= 3, N
    
    assert Xp.shape == (N, 3), Xp.shape
    assert q.shape == (M, N, 2), q.shape
    assert K.shape == (3, 3), K.shape

    # solve PnP for each mirror to get the mirrored 3D pts
    Cp = []
    if N > 3:
        for i in range(M):
            # Use non-negative observations only
            #mask = q[i,:,0] >= 0
            mask = np.isfinite(q[i,:,0])
            mN = np.sum(mask)
            mXp = Xp[mask,:]
            mqi = q[i,mask,:]
            #print(mask.shape, mXp.shape, mqi.shape)
            flags = flags if flags is not None else cv2.SOLVEPNP_ITERATIVE
            retval, rvec, tvec = cv2.solvePnP(mXp, np.ascontiguousarray(mqi).reshape((mN, 1, 2)), K, None, flags=flags)
            assert retval is True
            Cp.append(w2c(Xp, rvec, tvec))  # FIXME: mask 3D points too
    else:
        #assert False, "P3P mode is not implemented yet"
        Cp_candidates = []
        for i in range(M):
            # P3P fails ...
            flags = flags if flags is not None else cv2.SOLVEPNP_P3P
            retval, rvecs, tvecs = cv2.solveP3P(Xp, np.ascontiguousarray(q[i]).reshape((N, 1, 2)), K, None, flags=flags)
            #assert retval == 0, retval
            tmp = []
            for rv, tv in zip(rvecs, tvecs):
                tmp.append(w2c(Xp, rv, tv))
            Cp_candidates.append(tmp)
        Cp = sub_tmn_orth(np.array(Cp_candidates))
    Cp = np.array(Cp)

    # compute the extrinsic camera parameter with the proposed method
    R, T, n, d = sub_tnm_rt(Xp, Cp)  # FIXME: this part should also consider the observation mask

    # compute the reprojection error per pixel
    rep = sub_reproj(Xp, q, R, T, n, d, K)

    return R, T, n, d, rep

def cart2sph(xyz):
    xyz = xyz.reshape((-1, 3))
    theta = np.arccos(xyz[:,2])
    phi = np.sign(xyz[:,1])*np.arccos(xyz[:,0] / np.linalg.norm(xyz[:,:2], axis=1))
    return theta, phi

def sph2cart(theta, phi):
    st = np.sin(theta)
    ct = np.cos(theta)
    sp = np.sin(phi)
    cp = np.cos(phi)
    return np.array([st*cp, st*sp, ct]).T

def tnm_ba(Xp: np.ndarray, q: np.ndarray, K: np.ndarray, R0: np.ndarray, T0: np.ndarray, n0: np.ndarray, d0: np.ndarray, *, verbose=0) -> [np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Optimizes the camera and the mirror poses non-linearly

    Parameters
    ----------
    Xp : numpy.ndarray
        Nx3 matrix representing 3D coordinates of N reference points in the reference coordinate system.
    q : numpy.ndarray
        MxNx2 matrix representing 2D projections of N reference points obeserved via M mirrors.
        NaN values indicate "masked / invalid" points.
    K : numpy.ndarray
        3x3 intrinsic camera parameter.
    R0 : numpy.ndarray
        Initial estimate of 3x3 rotation matrix of the camera.
    T0 : numpy.ndarray
        Initial estimate of 3x1 translation vector of the camera.
    n0 : numpy.ndarray
        Initial estimate of Mx3 normal vectors of the mirror planes.
    d0 : numpy.ndarray
        Initial estimate of Mx1 distances between the camera and the mirrors.

    Returns
    -------
    R : numpy.ndarray
        Optimized 3x3 rotation matrix of the camera.
    T : numpy.ndarray
        Optimized 3x1 translation vector of the camera.
    n : numpy.ndarray
        Optimized Mx3 normal vectors of the mirror planes.
    d : numpy.ndarray
        Optimized Mx1 distances between the camera and the mirrors.
    rep : float
        The average of reprojection error per point
    """

    num_points = len(Xp)
    num_mirrors = len(n0)
    assert Xp.shape == (num_points, 3), Xp.shape
    assert q.shape == (num_mirrors, num_points, 2), q.shape
    assert R0.shape == (3, 3), R0.shape
    assert T0.shape == (3,) or T0.shape == (3, 1), T0.shape
    assert n0.shape == (num_mirrors, 3) or n0.shape == (num_mirrors, 3, 1), n0.shape
    assert d0.shape == (num_mirrors,), d0.shape

    def encode(R, T, n, d):
        x = []
        x.extend(cv2.Rodrigues(R)[0].flatten().tolist())
        x.extend(T.flatten().tolist())
        theta, phi = cart2sph(n)
        x.extend(theta.tolist())
        x.extend(phi.tolist())
        #x.extend(n.flatten().tolist())
        x.extend(d.flatten().tolist())
        return np.array(x)

    def decode(x):
        R = cv2.Rodrigues(x[:3])[0]
        T = x[3:6]
        m = (len(x)-6) // 3
        assert 3*m+6 == len(x)
        theta = x[6:6+m]
        phi = x[6+m:6+2*m]
        d = x[6+2*m:]
        n = sph2cart(theta, phi)
        return R, T, n, d

    def objfun(x, Xp, q, K):
        R_, T_, n_, d_ = decode(x)
        e = sub_reproj_vec(Xp, q, R_, T_, n_, d_, K).flatten()
        return e[~np.isnan(e)]

    x0 = encode(R0, T0, n0, d0)
    rep0 = sub_reproj(Xp, q, R0, T0, n0, d0, K)

    # double-check
    R_, T_, n_, d_ = decode(x0)
    assert np.allclose(R0, R_)
    assert np.allclose(T0, T_)
    assert np.allclose(n0, n_)
    assert np.allclose(d0, d_)

    ret = sp.optimize.least_squares(objfun, x0, args=(Xp, q, K), verbose=verbose)

    R, T, n, d = decode(ret['x'])
    rep = sub_reproj(Xp, q, R, T, n, d, K)

    return R, T, n, d, rep

"""
def 
# fill missing corners by dummy
all_obj_pts = []
all_img_pts = []
all_frames = pd.DataFrame({'frame': ALL_PTS['frame'].unique()}).set_index('frame')

for idx, p in enumerate(OBJ_PTS_PX):
    observation = ALL_PTS.loc[ALL_PTS['id'] == idx]
    if len(observation) == 0:
        print(f'Skipping corner {idx} with {len(observation)} observations.')
        continue

    # fill missing frames by (-1, -1)
    uv = all_frames.merge(observation, on='frame', how='left').fillna(-1)
    all_obj_pts.append(p)
    all_img_pts.append(uv[['u','v']].to_numpy())
"""


import unittest
from testbook import testbook

import numpy as np
import os
import cv2
import pycalib

# https://qiita.com/Syuparn/items/fbb8533e5f4fa3181ffd
def local_chdir(func):
    def _inner(*args, **kwargs):
        # get the current dir before entering the func
        dir_original = os.getcwd()
        # run func
        ret = func(*args, **kwargs)
        # back to the original dir
        os.chdir(dir_original)
        return ret
    return _inner

class TestPyCalib(unittest.TestCase):
    """
    verifies that all notebooks run without error
    """

    @classmethod
    def setUpClass(cls):
        pass

    @local_chdir
    def run_ipynb(self, file):
        os.chdir('./ipynb')
        with testbook(file, execute=True, timeout=600) as tb:
            pass

    def test_absolute_orientation(self):
        self.run_ipynb('absolute_orientation.ipynb')

    def test_absolute_orientation(self):
        self.run_ipynb('aruco_movie.ipynb')

    def test_calib2d3d(self):
        self.run_ipynb('calib2d3d.ipynb')

    def test_charuco_diamond(self):
        self.run_ipynb('charuco_diamond.ipynb')

    def test_example_gopro_step1_incalib(self):
        self.run_ipynb('example_gopro_step1_incalib.ipynb')

    def test_example_gopro_step2_kp(self):
        self.run_ipynb('example_gopro_step2_kp.ipynb')

    def test_example_gopro_step3_ba(self):
        self.run_ipynb('example_gopro_step3_ba.ipynb')

    def test_example_gopro_step4_floor(self):
        self.run_ipynb('example_gopro_step4_floor.ipynb')

    def test_excalib_2d(self):
        self.run_ipynb('excalib_2d.ipynb')

    def test_excalib_chess(self):
        self.run_ipynb('excalib_chess.ipynb')

    def test_incalib_charuco(self):
        self.run_ipynb('incalib_charuco.ipynb')

    def test_incalib_chess(self):
        self.run_ipynb('incalib_chess.ipynb')

    def test_ncam_ba(self):
        self.run_ipynb('ncam_ba.ipynb')

    def test_ncam_registration(self):
        self.run_ipynb('ncam_registration.ipynb')

    def test_ncam_triangulate(self):
        self.run_ipynb('ncam_triangulate.ipynb')

    def test_ncam_triangulate_consensus(self):
        self.run_ipynb('ncam_triangulate_consensus.ipynb')

    def test_excalib_charuco(self):
        self.run_ipynb('excalib_charuco.ipynb')

    def test_sphere(self):
        self.run_ipynb('sphere.ipynb')

    def test_qrtimecode(self):
        self.run_ipynb('qrtimecode.ipynb')

    def test_mirror(self):
        self.run_ipynb('mirror.ipynb')

    def test_mirror_charuco(self):
        self.run_ipynb('mirror_charuco.ipynb')

    def test_circle3d(self):
        self.run_ipynb('circle3d.ipynb')

    def test_aruco_movie(self):
        self.run_ipynb('aruco_movie.ipynb')


import numpy as np
import itertools
import cv2
import pycalib

def solve_circle_3d(K, ellipse, *, radius=1):
    """
    Returns the 2 solutions of the 3D circle center and normal vector in the camera coordinate system from its projection, i.e., the ellipse, by Chen et al. 2004.

    Parameters
    ----------
    K : ndarray
        Camera matrix
    ellipse : tuple
        (center, axes, angle) of the ellipse, given by cv2.fitEllipse()

    Returns
    -------
    c : ndarray, 2x3
        Center of the circle in the camera coordinate system
    n : ndarray, 2x3
        Normal vector of the plane of the circle in the camera coordinate system
    """

    assert K.shape == (3, 3), 'K must be a 3x3 camera matrix'
    assert len(ellipse) == 3, 'ellipse must be a tuple (center, axes, angle), given by cv2.fitEllipse()'
    assert radius > 0, 'radius must be positive'

    A, B, C, D, E, F = ellipse_to_quadric(ellipse, K[0, 2], K[1, 2])
    c, n = solve_circle_from_quadric(A, B, C, D, E, F, K[0, 0], radius=radius)

    return c, n

def solve_circle_from_quadric(A, B, C, D, E, F, focal_length, *, radius=1):
    """
    Returns the 2 solutions of the 3D circle center and normal vector in the camera coordinate system from the quadric equation of the ellipse, by Chen et al. 2004.

    Parameters
    ----------
    A : float
        Coefficient of x^2
    B : float
        Coefficient of xy
    C : float
        Coefficient of y^2
    D : float
        Coefficient of x
    E : float
        Coefficient of y
    F : float
        Constant term
    focal_length : float
        Focal length of the camera
    radius : float
        Radius of the circle

    Returns
    -------
    c : ndarray, 2x3
        Center of the circle in the camera coordinate system
    n : ndarray, 2x3
        Normal vector of the plane of the circle in the camera coordinate system
    """
    # Eq (5)
    Q = np.array([[A, B, D/focal_length], [B, C, E/focal_length], [D/focal_length, E/focal_length, F/focal_length/focal_length]])
    assert np.allclose(Q, Q.T)

    # Eq (4), for debugging
    #p = np.zeros((p_gt.shape[0], 3))
    #p[:,:2] = p_gt - K_gt[:2, 2]
    #p[:, 2] = focal_length
    #e = np.einsum('ij,ji->i', p, Q @ p.T)
    #assert np.allclose(e, 0, atol=1e-4)

    # Eq (11)
    eigvals, eigvecs = np.linalg.eigh(Q)
    assert np.allclose(Q, eigvecs @ np.diag(eigvals) @ eigvecs.T)
    #print(f'Eigenvalues: {eigvals}')
    #print(f'Eigenvectors:\n{eigvecs}')

    # Eq (16)
    for i0, i1, i2 in itertools.permutations([0, 1, 2]):
        l1, l2, l3 = eigvals[i0], eigvals[i1], eigvals[i2]
        if l1 * l2 > 0 and l1 * l3 < 0 and np.abs(l1) >= np.abs(l2):
            eigvals = np.array([l1, l2, l3])
            eigvecs = eigvecs[:, [i0, i1, i2]]
            break
    else:
        assert False
    assert np.allclose(Q, eigvecs @ np.diag(eigvals) @ eigvecs.T)

    # Eq (18)
    l1, l2, l3 = eigvals
    g = np.sqrt((l2-l3)/(l1-l3))
    h = np.sqrt((l1-l2)/(l1-l3))

    # Eqs (19), (20)
    c_candidates = []
    n_candidates = []

    for S1 in [-1, 1]:
        for S2 in [-1, 1]:
            for S3 in [-1, 1]:
                z0 = S3 * l2 * radius / np.sqrt(-l1*l3)
                x0 = - S2 * np.sqrt((l1-l2)*(l2-l3)) / l2 * z0
                y0 = 0
                c = z0 * eigvecs @ np.array([S2*l3/l2*h, 0, -S1*l1/l2*g]).reshape((3,1))
                n = eigvecs @ np.array([S2*h, 0, -S1*g]).reshape((3,1))
                R = np.array([[g, 0, S2*h], [0, -S1, 0], [S1*S2*h, 0, -S1*g]])
                if n[2] < 0 and c[2] > 0:
                    c_candidates.append(c)
                    n_candidates.append(n)

    assert len(c_candidates) == 2
    assert len(n_candidates) == 2

    return c_candidates, n_candidates

def draw_ellipse_quadric(img, u0, v0, A, B, C, D, E, F, color=(0, 255, 0), thickness=-1):
    """
    Draws an ellipse on the image using the quadric equation

    Parameters
    ----------
    img : ndarray
        Image
    u0 : float
        x-coordinate of the image center (principal point)
    v0 : float
        y-coordinate of the image center (principal point)
    A : float
        Coefficient of x^2
    B : float
        Coefficient of xy
    C : float
        Coefficient of y^2
    D : float
        Coefficient of x
    E : float
        Coefficient of y
    F : float
        Constant term
    color : tuple
        Color of the ellipse
    thickness : int
        Thickness of the ellipse (-1: filled)
    """

    w = img.shape[1]
    h = img.shape[0]
    xx, yy = np.meshgrid(np.arange(w), np.arange(h))
    x = xx.flatten() - u0
    y = yy.flatten() - v0
    e = A*x**2 + 2*B*x*y + C*y**2 + 2*D*x + 2*E*y + F
    e = e.reshape((h, w))
    if thickness<0:
        img[e < 0] = color
    else:
        img[np.logical_and(e<=thickness, e>=thickness)] = color
    return img

def ellipse_to_quadric(ellipse, u0, v0):
    """
    Returns A, B, C, D, E, F of the quadric equation of the ellipse

    Parameters
    ----------
    ellipse : tuple
        (center, axes, angle) of the ellipse, given by cv2.fitEllipse()
    u0 : float
        x-coordinate of the image center (principal point)
    v0 : float
        y-coordinate of the image center (principal point)

    Returns
    -------
    A : float
        Coefficient of x^2
    B : float
        Coefficient of xy
    C : float
        Coefficient of y^2
    D : float
        Coefficient of x
    E : float
        Coefficient of y
    F : float
        Constant term
    """

    u = ellipse[0][0] - u0
    v = ellipse[0][1] - v0
    a2 = (ellipse[1][1]/2)**2
    b2 = (ellipse[1][0]/2)**2
    theta = ellipse[2] / 180 * np.pi + np.pi/2
    ct = np.cos(theta)
    st = np.sin(theta)
    ct2 = ct*ct
    st2 = st*st

    A = ct2/a2 + st2/b2
    B = ct * st * (1/a2 - 1/b2)
    C = st2/a2 + ct2/b2
    D = - A*u - B*v
    E = - B*u - C*v
    F = A*u*u + 2*B*u*v + C*v*v - 1

    return A, B, C, D, E, F



import unittest
import os
import numpy as np
import bz2
import pycalib
import time
from scipy.optimize import least_squares

class TestPyCalibBa(unittest.TestCase):
    BAL_FILENAME = 'problem-49-7776-pre.txt.bz2'
    BAL_URL = 'https://grail.cs.washington.edu/projects/bal/data/ladybug/problem-49-7776-pre.txt.bz2'

    def test_bal(self):
        if not os.path.isfile(self.BAL_FILENAME):
            import urllib.request
            urllib.request.urlretrieve(self.BAL_URL, self.BAL_FILENAME)

        with bz2.open(self.BAL_FILENAME) as fp:
            camera_params, points_3d, camera_indices, point_indices, points_2d = pycalib.bal.bal_read(fp)

        print('flipping UV')
        points_2d = pycalib.bal.bal_flip_uv(points_2d)
        camera_params, mask = pycalib.bal.bal_cam9_to_cam17(camera_params)

        # allow k1, k2, p1, p2, k3
        mask[9:14] = True

        print(mask, len(mask))

        cam_opt, X_opt, e, ret = pycalib.ba.bundle_adjustment(camera_params, points_3d, camera_indices, point_indices, points_2d, mask=mask)

        print(f'cost = {ret.cost}')
        print(f'reprojection error = {e}')
        self.assertLess(ret.cost, 9963)
        self.assertLess(e, 0.7062)

    def test_sqrt(self):
        N = 5
        for i in range(100):
            A = np.random.randn(N, 2, 2)
            #ic(A)
            #ic(np.transpose(A, (0,2,1)))
            #a = A[1]
            #ic(a)
            #ic(a @ a.T)
            A = np.einsum('nij,njk->nik', A, np.transpose(A, (0,2,1)))
            #ic(A)
            self.assertTrue(np.all(np.linalg.det(A)>=0), 'A must be PSD matrices')
            B = pycalib.ba.sqrt_symmetric_2x2_mat(A)
            #ic(B)
            BB = np.einsum('nij,njk->nik', B, B)
            #ic(BB)
            np.testing.assert_allclose(A, BB)

            iB = np.linalg.inv(B)
            self.assertEqual(iB.shape, (N, 2, 2))
            BiB = np.einsum('nij,njk->nik', B, iB)
            np.testing.assert_allclose(BiB, np.tile(np.eye(2), (N, 1, 1)), atol=1e-6, rtol=1e-6)


if __name__ == '__main__':
    unittest.main()



import sys
import cv2
import numpy as np
import pycalib

def bal_triangulate(camera_params, camera_indices, point_indices, points_2d):
    num_cameras = len(camera_params)
    assert camera_params.ndim == 2
    pycalib.util.check_observations(camera_indices, point_indices, points_2d)


    if camera_params.shape[1] ==9:
        # BAL -> full camera parameters
        c = []
        for i in camera_params:
            c.append( bal_cam9_to_cam17(i) )
        camera_params = np.array(c)
    assert camera_params.shape[1] == 17


    K = []
    D = []
    P = []
    for c in camera_params:
        r, t, k, d= pycalib.ba.decode_camera_param(c)
        K.append(k)
        D.append(d)
        P.append(k @ np.hstack((r, t)))
    K = np.array(K)
    D = np.array(D)
    P = np.array(P)

    Y_est, PIDs_ok = pycalib.calib.triangulateN(K, D, P, camera_indices, point_indices, points_2d)

    return Y_est.T

def bal_recalib(camera_params, camera_indices, point_indices, points_2d):
    num_cameras = len(camera_params)
    assert camera_params.ndim == 2
    pycalib.util.check_observations(camera_indices, point_indices, points_2d)


    if camera_params.shape[1] ==9:
        # BAL -> full camera parameters
        c = []
        for i in camera_params:
            c.append( bal_cam9_to_cam17(i) )
        camera_params = np.array(c)
    assert camera_params.shape[1] == 17


    K = []
    D = []
    for c in camera_params:
        _, _, k, d= pycalib.ba.decode_camera_param(c)
        K.append(k)
        D.append(d)
    K = np.array(K)
    D = np.array(D)


    R, t, X, _ = pycalib.calib.excalibN(K, D, camera_indices, point_indices, points_2d)
    assert len(R) == num_cameras
    assert len(t) == num_cameras

    new_cameras = camera_params.copy()
    for i in range(num_cameras):
        new_cameras[i, 0:3] = cv2.Rodrigues(R[i])[0].reshape(-1)
        new_cameras[i, 3:6] = t[i].reshape(-1)

    return new_cameras, X

def bal_write(file, n_cameras, *, camera_params=None, points_3d=None, camera_indices=None, point_indices=None, points_2d=None):
    assert all( [camera_indices is None, point_indices is None, points_2d is None] ) or all( [camera_indices is not None, point_indices is not None, points_2d is not None] )

    n_observations = 0
    if camera_indices is not None:
        n_observations = len(camera_indices)
        assert n_observations == len(point_indices)
        assert points_2d.shape == (n_observations, 2)
    
    n_points = 0
    if points_3d is not None:
        n_points = len(points_3d)
        assert points_3d.shape == (n_points, 3)
    elif n_observations > 0:
        n_points = len(np.unique(point_indices))
        points_3d = np.zeros((n_points, 3))

    if camera_params is None:
        camera_params = np.zeros((n_cameras, 9))

    file.write(f'{n_cameras} {n_points} {n_observations}\n'.encode('utf-8'))

    if n_observations > 0:
        for c, p, uv in zip(camera_indices, point_indices, points_2d):
            file.write(f'{c} {p} {uv[0]} {uv[1]}\n'.encode('utf-8'))

    for i in camera_params.flatten():
        file.write(f'{i}\n'.encode('utf-8'))

    if n_points > 0:
        for i in points_3d.flatten():
            file.write(f'{i}\n'.encode('utf-8'))

def bal_info(file):
    fp = pycalib.util.open_z(file, 'r')
    n_cameras, n_points, n_observations = map(int, fp.readline().split())
    return n_cameras, n_points, n_observations

# https://scipy-cookbook.readthedocs.io/items/bundle_adjustment.html
def bal_read(file, *, verify_correspondences=True, verify_indices=True, n_camparams=9):
    """
    read BAL data as-is without applying any conversions
    """

    n_cameras, n_points, n_observations = map(
        int, file.readline().split())

    camera_indices = np.empty(n_observations, dtype=int)
    point_indices = np.empty(n_observations, dtype=int)
    points_2d = np.empty((n_observations, 2))

    for i in range(n_observations):
        camera_index, point_index, x, y = file.readline().split()
        camera_indices[i] = int(float(camera_index))
        point_indices[i] = int(float(point_index))
        points_2d[i] = [float(x), float(y)]

    camera_params = np.empty(n_cameras * n_camparams)
    for i in range(n_cameras * n_camparams):
        camera_params[i] = float(file.readline())
    camera_params = camera_params.reshape((n_cameras, -1))

    points_3d = np.empty(n_points * 3)
    for i in range(n_points * 3):
        points_3d[i] = float(file.readline())
    points_3d = points_3d.reshape((n_points, -1))

    if verify_indices:
        assert np.all(camera_indices >= 0)
        assert np.all(camera_indices < n_cameras)
        assert np.all(point_indices >= 0)
        assert np.all(point_indices < n_points)

    if verify_correspondences:
        pid, count = np.unique(point_indices, return_counts=True)
        assert count.min() >= 2, 'some points are not observed by 2 cameras'

    camera_indices = camera_indices.astype(np.int32)
    point_indices = point_indices.astype(np.int32)

    return camera_params, points_3d, camera_indices, point_indices, points_2d

def bal_flip_uv(points_2d):
    return - points_2d

# def bal_load_numpy(fp, *, use_initial_pose=True, need_uv_flip=True):
#     """
#     read BAL data with applying conversions
#     """
# 
#     # http://grail.cs.washington.edu/projects/bal/
# 
#     camera_params, points_3d, camera_indices, point_indices, points_2d = read_bal_data(fp)
# 
#     # Do we need to flip u,v directions?  (original BAL requires this since its projection model is p = -P / P.z)
#     if need_uv_flip:
#         #print('flipping UV', file=sys.stderr)
#         points_2d = - points_2d
# 
#     # Do we guess the initial pose by ourselves?
#     if use_initial_pose is False:
#         camera_params, points_3d = bal_recalib(camera_params, camera_indices, point_indices, points_2d)
# 
#     return camera_params, points_3d, camera_indices, point_indices, points_2d

def bal_cam9_to_cam17(camera_params):
    """ converts cameras with 9 params (r, t, f, k1, k2) to 17 params (r, t, f, cx, cy, k1, k2, p1, p2, k3, k4, k5, k6) """
    n = camera_params.shape[0]
    c, m = bal_cam9_to_cam14(camera_params)
    c = np.hstack( [c, np.zeros((n, 3))] )
    m = np.concatenate( [m, [False, False, False]] )
    return c, m

def bal_cam9_to_cam14(camera_params):
    """ converts cameras with 9 params (r, t, f, k1, k2) to 14 params (r, t, f, cx, cy, k1, k2, p1, p2, k3) """
    n = camera_params.shape[0]
    assert camera_params.shape[1] == 9

    c = np.zeros((n, 14))
    m = np.ones(14, dtype=bool)

    c[:, :6] = camera_params[:, :6] # r, t
    c[:, 6] = camera_params[:, 6] # f
    m[7:9] = False # cx, cy
    c[:, 9:11] = camera_params[:, 7:9] # k1, k2
    m[11:14] = False # p1, p2, k3

    return c, m




import unittest
import numpy as np
import cv2
import pycalib

from pycalib.calib import absolute_orientation

class TestPyCalibAo(unittest.TestCase):
    def test_ao(self):
        # synthetic data
        N = 10
        P = np.random.random((3, N))
        R, _ = cv2.Rodrigues(np.random.random(3))
        t = np.random.random(3)
        s = np.random.random(1)
        Q = s * (R @ P) + t[:, None]

        # solve
        R2, t2, s2 = absolute_orientation(P, Q)
        P2 = s2 * (R2 @ P) + t2[:, None]

        # check
        np.testing.assert_allclose(R, R2)
        np.testing.assert_allclose(t, t2)
        np.testing.assert_allclose(s, s2)

    def test_ao2(self):
        # synthetic data
        N = 10
        P = np.random.random((3, N))
        R, _ = cv2.Rodrigues(np.random.random(3))
        t = np.random.random(3)
        s = 1
        Q = s * (R @ P) + t[:, None]

        # solve
        R2, t2, s2 = absolute_orientation(P, Q, no_scaling=True)
        P2 = s2 * (R2 @ P) + t2[:, None]

        # check
        np.testing.assert_allclose(R, R2)
        np.testing.assert_allclose(t, t2)
        np.testing.assert_allclose(s, s2)




import cv2
import numpy as np
from scipy.sparse import lil_matrix
from scipy.optimize import least_squares

import matplotlib.pyplot as plt

import pycalib.plot
import pycalib.calib

# https://scipy-cookbook.readthedocs.io/items/bundle_adjustment.html

def sqrt_symmetric_2x2_mat(m_Nx2x2):
    """ Compute the sqrt of symmetric and positive semi-definite 2x2 matrices

    https://en.wikipedia.org/wiki/Square_root_of_a_2_by_2_matrix

    Parameters
    ----------
    m_Nx2x2 : ndarray
        a N x 2 x 2 array (each 2x2 matrix must be symmetric and positive semi-definite)

    Returns
    -------
    sq_Nx2x2 : ndarray
        the sqrt of m_Nx2x2
    """

    N = m_Nx2x2.shape[0]
    assert m_Nx2x2.shape == (N,2,2)
    assert np.allclose(m_Nx2x2, np.transpose(m_Nx2x2, axes=(0,2,1))), "m_Nx2x2 must be symmetric"

    s = np.sqrt(np.linalg.det(m_Nx2x2))
    t = np.sqrt(np.trace(m_Nx2x2, axis1=1, axis2=2) + 2*s)
    assert s.shape == (N,)
    assert t.shape == (N,)
    if np.any(t == 0):
        print('[WARN] Some of covariance matrices are not positive definite.')
        t[t==0] = 0.1 # fixme
    sq = (1/t).reshape((-1,1,1)) * (m_Nx2x2 + s.reshape((-1,1,1))*np.eye(2))
    assert sq.shape == (N, 2, 2)

    return sq

def rotate(points, rot_vecs):
    """Rotate points by given rotation vectors.

    Rodrigues' rotation formula is used.
    see also https://github.com/strasdat/Sophus/blob/master/sophus/so3.hpp#L257
    """
    theta = np.linalg.norm(rot_vecs, axis=1)[:, np.newaxis]
    with np.errstate(invalid='ignore'):
        v = rot_vecs / theta
        v = np.nan_to_num(v)
    dot = np.sum(points * v, axis=1)[:, np.newaxis]
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    return cos_theta * points + sin_theta * np.cross(v, points) + dot * (1 - cos_theta) * v

def project(points, camera_params):
    """Convert 3-D points to 2-D by projecting onto images."""

    assert camera_params.shape[1] == 17

    points_proj = rotate(points, camera_params[:, :3])
    points_proj += camera_params[:, 3:6]
    points_proj = points_proj[:, :2] / points_proj[:, 2, np.newaxis]

    fx = camera_params[:, 6]
    fy = camera_params[:, 6]
    cx = camera_params[:, 7]
    cy = camera_params[:, 8]
    k1 = camera_params[:, 9]
    k2 = camera_params[:, 10]
    p1 = camera_params[:, 11]
    p2 = camera_params[:, 12]
    k3 = camera_params[:, 13]
    k4 = camera_params[:, 14]
    k5 = camera_params[:, 15]
    k6 = camera_params[:, 16]

    n = np.sum(points_proj**2, axis=1)
    n2 = n**2
    n3 = n2 * n
    kdist = (1 + k1 * n + k2 * n2 + k3 * n3) / (1 + k4 * n + k5 * n2 + k6 * n3)
    pdist = 2 * points_proj[:,0] * points_proj[:,1]
    u = points_proj[:,0]*kdist + p1*pdist + p2*(n2+2*points_proj[:,0]**2)
    v = points_proj[:,1]*kdist + p2*pdist + p1*(n2+2*points_proj[:,0]**2)

    points_proj[:, 0] = fx*u + cx
    points_proj[:, 1] = fy*v + cy

    return points_proj

def x2c(x, c0, mask):
    # base params
    camera_params = c0.copy()
    n_cameras = len(camera_params)
    assert c0.shape == (n_cameras, 17)

    if mask.shape == (17,):
        # 1D == shared mask
        n_params = np.sum(mask)
        n_params_total = n_cameras * n_params
        camera_params[:, mask] = x[:n_params_total].reshape((n_cameras, n_params))
    else:
        # 2D == camera-wise mask
        assert mask.shape == (n_cameras, 17)

        if mask.dtype == bool:
            n_params_total = np.sum(mask != 0)
            camera_params = camera_params.flatten()
            camera_params[mask.flatten()] = x[:n_params_total]
            camera_params = camera_params.reshape((n_cameras, 17))
        else:
            assert mask.dtype == int
            uval, uval_indices, uval_inverse, uval_counts = np.unique(mask, return_index=True, return_inverse=True, return_counts=True)
            idx = uval_inverse.reshape(mask.shape)
            camera_params[mask != 0] = x[uval_inverse[mask.flatten() != 0]-1]
            n_params_total = np.sum(uval != 0)

    assert camera_params.shape == (n_cameras, 17)
    return camera_params, n_params_total

def c2x(c, mask):
    assert c.shape == mask.shape

    if mask.dtype == bool:
        x = c[mask != 0].flatten()
    else:
        assert mask.dtype == int
        uval, uval_indices = np.unique(mask, return_index=True)
        x = c.flatten()[uval_indices[np.where(uval != 0)[0]]]

    return x

def reprojection_error(params, n_cameras, n_points, camera_indices, point_indices, points_2d, mask, weights, inv_stdev, cam0):
    """Compute residuals.

    `params` contains camera parameters and 3-D coordinates.
    """
    camera_params, n_params = x2c(params, cam0, mask)
    points_3d = params[n_params:].reshape((n_points, 3))
    points_proj = project(points_3d[point_indices], camera_params[camera_indices])
    e = points_proj - points_2d
    if inv_stdev is not None:
        e = np.einsum('kij,kj->ki', inv_stdev, e)
    return (e * weights[:,None]).ravel()

def bundle_adjustment_sparsity(n_cameras, n_points, camera_indices, point_indices, mask):
    # returns a m-by-n binary matrix
    # m: observations (num of 2d point x 2 (= u, v))
    # n: parameters (camera params + num of 3d points x 3)

    assert len(mask.shape) == 2
    assert mask.dtype == int

    # find the mask boundary per camera
    uval, uval_indices, uval_inverse, uval_counts = np.unique(mask, return_index=True, return_inverse=True, return_counts=True)
    mask_idx = uval_inverse.reshape(mask.shape)

    n_params = np.sum(uval != 0)

    m = camera_indices.size * 2
    n = n_params + n_points * 3
    A = lil_matrix((m, n), dtype=int)

    # for each camera
    for i in range(len(mask)):
        idx = np.where(camera_indices == i)[0]
        u = np.unique(mask_idx[i,:])
        u = u[u != 0] - 1

        for j in u:
            A[2 * idx, j] = 1
            A[2 * idx+1, j] = 1

    i = np.arange(camera_indices.size)
    for s in range(3):
        A[2 * i, n_params + point_indices * 3 + s] = 1
        A[2 * i + 1, n_params + point_indices * 3 + s] = 1

    #print(A.toarray())
    return A

def bundle_adjustment(camera_params, points_3d, camera_indices, point_indices, points_2d, *, verbose=2, mask=None, weights=None, loss='linear', cov=None):
    """
    Optimize camera poses and intrinsics non-linearly

    Parameters
    ----------
    camera_params : ndarray
        C x 17 array of camera parameters (rvec, tvec, f, u0, v0, k1, k2, p1, p2, k3, k4, k5, k6)
    points_3d : ndarray
        M x 3 array of 3D points
    camera_indices : ndarray
        N array of camera indices. points_3d[point_indices[i]] is observed by camera_indices[i] at points_2d[i]
    point_indices : ndarray
        N array of point indices. points_3d[point_indices[i]] is observed by camera_indices[i] at points_2d[i]
    points_2d: ndarray
        N x 2 array of 2D points.  points_3d[point_indices[i]] is observed by camera_indices[i] at points_2d[i]
    mask: ndarray
        A bool array of length 17 or a Cx17 bool matrix to specify the camera parameters to optimize.
    weights:
        N array of weights to scale the error at each point.  Set 0 to ignore the point.
    verbose: int
        {0, 1, 2}.  See `verbose` arg of `scipy.optimize.least_squares()`.
    loss: str or callable
        `linear`, `soft_l1`, `huber`, `cauchy`, or `arctan`.  See `loss` arg of `scipy.optimize.least_squares()`.
    cov: ndarray
        N x 2 x 2 array of covariance matrices of 2D points.

    Returns
    -------
    camera_params : ndarray
        C x 17 array of camera parameters (rvec, tvec, f, u0, v0, k1, k2, p1, p2, k3, k4, k5, k6)
    points_3d : ndarray
        M x 3 array of 3D points
    reproj : float
        reprojection error (RMSE)
    res : OptimizeResult
        output of `scipy.optimize.least_squares()`


    ToDo
    ----
    - add bounds support
    - static 3D points
    - [x] shared camera intrinsic params == moving camera
    - [x] shared camera extrinsic params == zoom camera
    """

    n_cameras = camera_params.shape[0]
    n_points = points_3d.shape[0]
    n_observations = points_2d.shape[0]

    assert camera_params.shape[1] == 17, camera_params.shape[1]
    pycalib.util.check_observations(camera_indices, point_indices, points_2d)

    # mask is 2D int>=0
    if mask is None:
        mask = np.ones(camera_params.shape, dtype=bool)
    elif mask.shape == (17,):
        # 1D -> 2D
        mask = np.tile(mask, (len(camera_params), 1))
    if mask.dtype == bool:
        m = np.arange(len(mask.flatten())).reshape(mask.shape) + 1
        m[mask == 0] = 0
        mask = m
    assert mask.dtype == int
    assert mask.shape == camera_params.shape
    assert np.min(mask) >= 0

    # weight is 1D
    if weights is None:
        weights = np.ones(len(point_indices))
    assert weights.ndim == 1
    assert len(weights) == point_indices.shape[0]

    # cov is No x 2 x 2
    if cov is not None:
        assert cov.shape == (n_observations, 2, 2), f'{cov.shape} != {n_observations}x2x2'
        assert np.allclose(cov, np.transpose(cov, axes=(0,2,1))), "cov must be symmetric"

        # Standard deviation matrix == sqrt of covariance matrix
        # https://en.wikipedia.org/wiki/Standard_deviation#Standard_deviation_matrix
        stdev = sqrt_symmetric_2x2_mat(cov)
        inv_stdev = np.linalg.inv(stdev)
        assert inv_stdev.shape == (n_observations, 2, 2)
    else:
        inv_stdev = None

    camera_params0 = camera_params.copy()

    A = bundle_adjustment_sparsity(n_cameras, n_points, camera_indices, point_indices, mask)

    x0 = np.hstack((c2x(camera_params, mask), points_3d.ravel()))

    res = least_squares(reprojection_error, x0, jac_sparsity=A, verbose=verbose, x_scale='jac', ftol=1e-4, method='trf', loss=loss, args=(n_cameras, n_points, camera_indices, point_indices, points_2d, mask, weights, inv_stdev, camera_params0))

    camera_params, n = x2c(res.x, camera_params0, mask)

    points_3d = res.x[n:].reshape((-1, 3))

    e = reprojection_error(res.x, n_cameras, n_points, camera_indices, point_indices, points_2d, mask, weights, None, camera_params0).reshape((-1, 2))
    e = np.linalg.norm(e, axis=1)
    assert( len(e) == n_observations )
    e = np.sqrt(np.mean(e))

    return camera_params, points_3d, e , res

def encode_camera_params(R_Nx3x3, T_Nx3, K_Nx3x3, D_Nx8):
    N = len(R_Nx3x3)
    assert R_Nx3x3.shape == (N, 3, 3), R_Nx3x3.shape
    assert T_Nx3.shape == (N, 3, 1), T_Nx3.shape
    assert K_Nx3x3.shape == (N, 3, 3), K_Nx3x3.shape
    assert D_Nx8.shape == (N, 8), D_Nx8.shape

    p = []
    for r, t, k, d in zip(R_Nx3x3, T_Nx3, K_Nx3x3, D_Nx8):
        p.append( encode_camera_param(r, t, k, d) )
    p = np.array(p)

    assert p.shape == (N, 17)

    return p


def decode_camera_params(params_Nx17):
    N = len(params_Nx17)
    assert params_Nx17.shape == (N, 17)

    R = []
    T = []
    K = []
    D = []

    for p in params_Nx17:
        r, t, k, d = decode_camera_param(p)
        R.append(r)
        T.append(t)
        K.append(k)
        D.append(d)

    R = np.array(R)
    T = np.array(T)
    K = np.array(K)
    D = np.array(D)

    assert R.shape == (N, 3, 3), R.shape
    assert T.shape == (N, 3, 1), T.shape
    assert K.shape == (N, 3, 3), K.shape
    assert D.shape == (N, 8), D.shape

    return R, T, K, D


def encode_camera_param(rmat, tvec, K, distCoeffs):
    """
    Convert camera parameters for optimization (mat -> 17 params)
    """

    x = np.zeros(17)
    x[:3] = cv2.Rodrigues(rmat)[0].reshape(-1)
    x[3:6] = tvec.reshape(-1)
    x[6] = K[0,0]
    x[7] = K[0,2]
    x[8] = K[1,2]
    x[9:9+len(distCoeffs)] = distCoeffs
    return x

def decode_camera_param(x):
    """
    Convert optimized camera parameters (17 params -> mat)
    """

    rmat = cv2.Rodrigues(x[:3])[0]
    tvec = x[3:6].reshape((3, 1))
    K = np.eye(3)
    K[0,0] = x[6]
    K[1,1] = x[6]
    K[0,2] = x[7]
    K[1,2] = x[8]
    distCoeffs = x[9:17]
    return rmat, tvec, K, distCoeffs

def make_mask(r, t, *, f=False, u0=False, v0=False, k1=False, k2=False, p1=False, p2=False, k3=False, k4=False, k5=False, k6=False):
    """
    Generate a mask to indicate the parameters to be optimized

    Parameters
    ----------
    r : bool
        Optimize the rotation matrix if True
    t : bool
        Optimize the translation vector if True
    f : bool
        Optimize the focal length if True
    u0 : bool
        Optimize u0 if True
    v0 : bool
        Optimize v0 if True
    k1 : bool
        Optimize k1 if True
    k2 : bool
        Optimize k2 if True
    p1 : bool
        Optimize p1 if True
    p2 : bool
        Optimize p2 if True
    k3 : bool
        Optimize k3 if True
    k4 : bool
        Optimize k4 if True
    k5 : bool
        Optimize k5 if True
    k6 : bool
        Optimize k6 if True
    """

    r = np.tile(r, 3)
    t = np.tile(t, 3)
    return np.concatenate([r, t, [f, u0, v0, k1, k2, p1, p2, k3, k4, k5, k6]]).astype(bool)

"""
def excalib2_ba(p1, p2, A1, d1, A2, d2, *, verbose=0):
    R, t, E, status, X = pycalib.calib.excalib2(p1, p2, A1, d1, A2, d2)
    assert np.all(status == 1)

    # BA
    Nc = 2
    Np = X.shape[1]

    ## Camera parameters
    camera_params = np.array([
        encode_camera_param( np.eye(3), np.zeros((3,1)), A1, np.zeros(5) if d1 is None else d1 ),
        encode_camera_param( R,         t,               A2, np.zeros(5) if d2 is None else d2 ),
    ])

    ## camera_indices[i] == the camera observes point_2d[i,:]
    camera_indices = np.repeat(np.arange(Nc), Np)

    ## point_indices[i] == the 3D point behind point_2d[i,:]
    point_indices = np.tile(np.arange(Np), Nc)

    ## Optimization target
    ## R, t, f, u0, v0, k1, k2, p1, p2, k3, k4, k5, k6
    mask = make_mask(True, True, False, False, False, False, False, False, False, False)

    ## 3D est (Np, 3)
    X_est = X.T

    ## 2D est (Nc, Np, 2)
    p1 = pycalib.calib.transpose_to_col(p1, 2).astype(float)
    p2 = pycalib.calib.transpose_to_col(p2, 2).astype(float)
    x_est = np.array( [ p1, p2 ] )

    ## optim
    cam_opt, X_opt, ret = bundle_adjustment(camera_params, X_est, camera_indices, point_indices, x_est.reshape((-1, 2)), mask=mask, verbose=verbose)

    rmat1, tvec1, k1, d1 = decode_camera_param(cam_opt[0])
    rmat2, tvec2, k2, d2 = decode_camera_param(cam_opt[1])

    assert np.allclose(k1, A1)
    assert np.allclose(k2, A2)
    assert np.count_nonzero(d1) == 0
    assert np.count_nonzero(d2) == 0

    #c1 = R1 x + t1
    #c2 = R2 x + t2
    #x = R1.T @ (c1 - t1)
    #c2 = R2 R1.T (c1 - t1) + t2

    R = rmat2 @ rmat1.T
    t = tvec2 - R @ tvec1
    X = rmat1 @ X_opt.T  + tvec1
    E = pycalib.calib.skew(t) @ R

    return R, t, E, status, X
"""




import unittest
import numpy as np
import pycalib
import cv2

class TestPyCalibDistortion(unittest.TestCase):

    def test_distortion(self):
        np.random.seed(0)

        # N-cameras
        Nc = 15

        # dummy data
        K_all = []
        D_all = []
        R_all = []
        T_all = []
        P_all = []
        for c in range(Nc):
            W = 640 + np.random.normal(0, 10)
            H = 480 + np.random.normal(0, 10)
            fx = 320 + np.random.normal(0, 10)
            fy = fx
            u0 = W // 2 + np.random.normal(0, 10)
            v0 = H // 2 + np.random.normal(0, 10)
            K = np.array( [[ fx, 0, u0 ], [0, fy, v0], [0, 0, 1]])
            D = np.random.normal(0, 0.1, 8)
            tvec = np.random.normal(0, 1, 3)
            tvec = tvec / np.linalg.norm(tvec) * (3 + np.random.rand())
            R, T = pycalib.calib.lookat(tvec, np.zeros(3), np.array([0, 1, 0]))
            P = K @ np.hstack([R, T])

            K_all.append(K)
            D_all.append(D)
            R_all.append(R)
            T_all.append(T)
            P_all.append(P)

        K_all = np.array(K_all)
        D_all = np.array(D_all)
        R_all = np.array(R_all)
        T_all = np.array(T_all)
        P_all = np.array(P_all)

        # grid points in 3D
        X_3d = np.array(np.meshgrid(np.linspace(-1, 1, 3), np.linspace(-1, 1, 3), np.linspace(-1, 1, 3))).reshape((3, -1)).T  # 3D grid points
        Np = len(X_3d)

        # projections
        X_2d = []
        for K, D, R, T, P in zip(K_all, D_all, R_all, T_all, P_all):
            x, _ = cv2.projectPoints(X_3d, R, T, K, D)
            X_2d.append(x)
        X_2d = np.array(X_2d)
        self.assertEqual(X_2d.shape, (Nc, Np, 1, 2))

        # undistort points
        U = []
        for x, K, D, R, T, P in zip(X_2d, K_all, D_all, R_all, T_all, P_all):
            u = pycalib.calib.undistort_points(x.reshape((-1, 2)), K, D)
            #u = cv2.undistortPointsIter(x, K, D, np.eye(3), K, (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 1000, 1e-8))

            # n3d = cv2.convertPointsToHomogeneous(u)
            # n3d = np.linalg.inv(K) @ n3d.reshape((-1, 3)).T
            # y, _ = cv2.projectPoints(n3d.T, np.eye(3), np.zeros(3), K, D)
            # print(y)
            d = pycalib.calib.distort_points(u.reshape((-1,1,2)), K, D)
            np.testing.assert_allclose(x, d, rtol=1e-6, atol=1e-6)
            U.append(u)
        U = np.array(U).reshape((Nc, -1, 2))

        # triangulate at once
        X = pycalib.calib.triangulate_Npts(U, P_all)
        np.testing.assert_allclose(X, X_3d, atol=1e-6)

        # triangulate one by one
        X = []
        for i in range(Np):
            x = pycalib.calib.triangulate(U[:,i,:], P_all)
            X.append(x)
        X = np.array(X)[:,:3]
        np.testing.assert_allclose(X, X_3d, atol=1e-6)


import numpy as np
import cv2
import dateutil.parser
from datetime import datetime
from tqdm import tqdm

def str2datetime_gopro(s):
    return datetime.strptime(s + '000', 'oT%y%m%d%H%M%S.%f')

def str2datetime_iso(s):
    return dateutil.parser.isoparse(s) 

def detect_qrcode_gopro(filename, *, begin=0, end=-1, fmt='gopro', verbose_lv=0):
    """ Detect QR code for each frame, and parse it as timestamp

    Use ./data/qrtimecode/gopro+js.html to show the QR code.

    Parameters
    ----------
    filename: str
        source video filename
    begin: int
        frame to start detection
    end: int
        frame to stop detection
    fmt: str
        either 'gopro' or 'iso'
    verbose_lv: int
        verbosity level

    Returns
    -------
    X: ndarray
        frame indices (int)
    Y: ndarray
        detected timestamps (float)
    FPS: float
        framerate
    """

    cap = cv2.VideoCapture(filename)
    qcd = cv2.QRCodeDetector()
    X = []
    Y = []
    
    str2datetime = str2datetime_gopro
    if fmt == 'iso':
        str2datetime = str2datetime_iso

    idx_max = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    if end < 0:
        end = idx_max
    else:
        end = min(end, idx_max)

    idx = 0

    pbar = tqdm(total=end-begin, disable=(verbose_lv==0))
    while True:
        ret, frame = cap.read()
        if frame is None:
            break

        if begin <= idx:
            ret_qr, decoded_info, points, _ = qcd.detectAndDecodeMulti(frame)
            if ret_qr:
                for s, p in zip(decoded_info, points):
                    if s:
                        ts = str2datetime(s)
                        X.append(idx)
                        Y.append(ts.timestamp())
                        if verbose_lv > 1:
                            print(f'{idx:12}, {ts}')
                pbar.set_description(f'Found in {len(X)} frames')
            pbar.update(1)
        idx += 1

        if idx >= end:
            break
    pbar.close()

    return np.array(X), np.array(Y), cap.get(cv2.CAP_PROP_FPS)


def estimate_offset_only(X_all, Y_all, alpha):
    X = X_all * alpha
    beta = np.mean(Y_all - X_all)
    return beta
    
def estimate_offset(X_all, Y_all, *, fps=0):
    """ Estimate time offsets between cameras from detected timestamps

    Parameters
    ----------
    X_all: list of np.ndarray
        frame indices of each timestamp
    Y_all: list of np.ndarray
        timestamps at frame indices given as X

    Returns
    -------
    offset: float
        frame offset between the cameras
    alpha: float
        slope alpha to map frame index to timestamp
    beta: ndarray
        camera-specific intercept beta to map frame index to timestamp
    rmse: ndarray
        root-mean-square residuals between Y and alpha*X+beta
    """

    if fps>0:
        alpha = 1.0 / fps
        beta = []
        for c, (X, Y) in enumerate(zip(X_all, Y_all)):
            b = estimate_offset_only(X, Y, alpha)
            beta.append(b)
        beta = np.array(beta)
    else:
        N = np.sum([ len(X) for X in X_all ])
        A = np.zeros((N, len(X_all)+1))
        b = np.zeros(N)
        
        curr = 0
        for c, (X, Y) in enumerate(zip(X_all, Y_all)):
            N = len(X)
            assert N == len(Y), f'len(X) != len(Y) for camera {c}'
    
            A[curr:curr+N, 0] = X
            A[curr:curr+N, 1+c] = 1
            b[curr:curr+N] = Y
            curr += N
        
        x = np.linalg.solve(A.T@A, A.T@b.reshape((-1,1)))
        x = x.flatten()
        alpha = x[0]
        beta = x[1:]

    offset = (beta - beta[0]) / alpha
    
    rmse_all = []
    for c, (X, Y) in enumerate(zip(X_all, Y_all)):
        Y_est = alpha * X + beta[c]
        e = np.sqrt(np.mean((Y-Y_est)**2))
        rmse_all.append(e)
        
    return offset, alpha, beta, np.array(rmse_all)



import numpy as np
import scipy.optimize
import cv2

# memo: cv2.fitEllipse returns ( (cx, cy), (2*a, 2*b), (theta_in_deg) )

def fit_ellipse(gray, *, min_sz=0, max_sz=-1, force_convex_hull=True):
    contours, _ = cv2.findContours(gray, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

    if max_sz < 0:
        max_sz = gray.shape[0]*gray.shape[1]

    for cnt in contours:
        M = cv2.moments(cnt)
        if M['m00'] < min_sz or M['m00'] > max_sz:
            continue
        if force_convex_hull:
            cnt = cv2.convexHull(cnt, returnPoints=True)
        if len(cnt) < 5:
            continue
        ellipse = cv2.fitEllipse(cnt)
        yield ellipse, cnt

def resample_ellipse(ellipse, N):
    t = np.linspace(0, 2*np.pi, N)
    p = np.array([ellipse[1][0] * np.cos(t), ellipse[1][1] * np.sin(t)]) / 2
    theta = ellipse[2] / 180 * np.pi
    rot = np.array([ [np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)] ])
    p = rot @ p + np.array(ellipse[0]).reshape((2, 1))
    return p.T.reshape((-1, 1, 2))


def render_sphere(center_3d, radius, K, img_w, img_h):
    # project a sphere to image (for generating dummy data)

    # [u, v, 1].T
    u, v = np.meshgrid(np.arange(img_w), np.arange(img_h))
    uv1 = np.dstack([u, v, np.ones(u.shape)])

    # (nx, ny, 1) = K^{-1} @ [u, v, 1].T
    nxyz = np.einsum('xy,ijy->ijx', np.linalg.inv(K), uv1)

    # nearest point on the viewing ray from the sphere center
    t = np.einsum('x,ijx->ij', center_3d, nxyz)
    t = t / np.einsum('ijx,ijx->ij', nxyz, nxyz)
    p = nxyz * t[:,:,None]

    # distances from the sphere center to each viewing ray
    d = np.linalg.norm(center_3d - p, axis=2)
    #print(np.max(d), np.min(d))

    # if the distance is smaller than the radius, it is in the sphere
    img = ((d <= radius) * 255).astype(np.uint8)

    return img


def calc_sphere_center_from_ellipse(ellipse, K, r=None):
    # Maalek and Lichti, "Correcting the Eccentricity Error of Projected Spherical Objectsin Perspective Cameras," 2021
    #ellipse = cv2.fitEllipse(cnt)
    cx, cy = ellipse[0]
    a, b = np.array(ellipse[1]) / 2
    theta = ellipse[2] / 180 * np.pi

    if a < b:
        a, b = b, a
        theta += np.pi / 2

    if cx == K[0,2] and cy == K[1,2]:
        return np.array([cx, cy])

    dir = np.array([cx-K[0,2], cy-K[1,2]])
    dir = dir / np.linalg.norm(dir)

    fe = np.sqrt(a**2 - b**2)
    ee = fe / np.sqrt(1 + (K[0,0]/b)**2)

    return np.array([cx, cy]) - ee * dir


def fit_sphere_center_3d_to_ellipse(cnt, K, *, r=1, resample=False):
    Kinv = np.linalg.inv(K)

    # x0 is an initial guess of the sphere center
    ellipse = cv2.fitEllipse(cnt)
    x0 = Kinv @ np.array([ellipse[0][0], ellipse[0][1], 1]).reshape((3,1))
    x0 *= K[0,0] / ((ellipse[1][0] + ellipse[1][1]) / 2) * r

    # resample?
    if resample:
        cnt = resample_ellipse(ellipse, 100)

    uv1 = np.dstack([ cnt[:,0,0], cnt[:,0,1], np.ones(cnt.shape[0])]).reshape((-1, 3))
    nxyz = np.einsum('xy,iy->ix', Kinv, uv1)
    nxyz = nxyz / np.linalg.norm(nxyz, axis=1)[:,None]
    r2 = r*r

    def f(x):
        # p is the closest point on the ray along nxyz to the sphere center x
        t = np.einsum('x,ix->i', x, nxyz)
        p = nxyz * t[:,None] - x
        # distance from x to p should be equal to the radius
        d = np.einsum('ij,ij->i', p, p) - r2
        return d

    def jac(x):
        t = np.einsum('x,ix->i', x, nxyz)
        p = nxyz * t[:,None]
        d = p - x

        a = np.einsum('ij,ik->ijk', nxyz, nxyz)
        a -= np.broadcast_to(np.eye(3), a.shape)
        a *= 2

        return (a @ d[:,:,None]).reshape(d.shape)

    ret = scipy.optimize.least_squares(f, x0.flatten(), jac, bounds=([-np.inf, -np.inf, 0], np.inf))
    return ret['x']





import cv2
import numpy as np
import pandas as pd
import pycalib
import multiprocessing
import itertools
from tqdm.auto import tqdm

def undistort_inplace(df, cid, camera_matrix, dist_coeffs, *, key_cid='camera', key_x='x', key_y='y'):
    """
    Undistort 2D points in DataFrame in place

    Parameters
    ----------
    df : pandas.DataFrame
        list of 2D points
    cid : int
        target camera id
    camera_matrix : ndarray
        camera matrix
    dist_coeffs : ndarray
        distortion coefficients
    key_cid : str
        column name of camera id
    key_x : str
        column name of x
    key_y : str
        column name of y
    """

    df_c = df[df[key_cid] == cid]
    if len(df_c) == 0:
        return
    pt2d = df_c[[key_x, key_y]].to_numpy()
    df.loc[df_c.index, [key_x, key_y]] = pycalib.calib.undistort_points(pt2d.reshape((-1, 1, 2)), camera_matrix, dist_coeffs).reshape((-1, 2))


def triangulate(df, P, cid1, cid2, *, reproj_th=10, primary_key=['frame', 'label'], key_cid='camera', key_x='x', key_y='y'):
    if 'index' in df:
        df = df.drop(columns='index')

    # Valid pairs == inner-join on frame and keypoint
    d = pd.merge(df[df[key_cid] == cid1], df[df[key_cid] == cid2], on=primary_key, suffixes=[f'_{cid1}', f'_{cid2}'])
    if len(d) == 0:
        raise Exception(f'No valid corresponding points between {cid1} and {cid2}')
    #print(f'Found {len(d)} valid corresponding points between {cid1} and {cid2}')

    # Triangulate
    p1 = d[[f'{key_x}_{cid1}', f'{key_y}_{cid1}']].to_numpy()
    p2 = d[[f'{key_x}_{cid2}', f'{key_y}_{cid2}']].to_numpy()
    X = cv2.triangulatePoints(P[cid1], P[cid2], p1.T, p2.T)
    X = X / X[3]

    # Reproject to all cameras
    x = np.einsum('cij,jp->cip', P, X)
    x = x / x[:,2:3,:]

    # Merge the other 2D observation by left-outer-join on frame and keypoint
    for c in range(len(P)):
        if c in [cid1, cid2]:
            continue
        d = pd.merge(d, df[df[key_cid] == c].rename(columns={key_cid:f'{key_cid}_{c}', key_x:f'{key_x}_{c}', key_y:f'{key_y}_{c}'}), on=primary_key, how='left')

    # Extract N x (x0, y0, x1, y1, ...)
    colnames = [x for c in range(len(P)) for x in [f'{key_x}_{c}', f'{key_y}_{c}']]
    y = d[colnames].to_numpy()

    # Reshape to N x C x 2
    y = y.reshape((-1, len(P), 2))

    # Transpose to C x 2 x N
    y = np.rollaxis(y, 0, 3)

    # Reprojection error
    e = x[:,:2,:] - y
    e = np.linalg.norm(e, axis=1)

    # Save the triangulated 3D points, inliers, and outliers
    d['X'] = X[0]
    d['Y'] = X[1]
    d['Z'] = X[2]

    d['reproj'] = list(map(lambda x : tuple(x), e.T))
    d['outliers'] = [tuple(np.where(i)[0]) for i in (e >= reproj_th).T]
    d['inliers'] = [tuple(np.where(i)[0]) for i in (e < reproj_th).T]
    d['n_outliers'] = np.sum(e >= reproj_th, axis=0).astype(int)
    d['n_inliers'] = np.sum(e < reproj_th, axis=0).astype(int)
    #d['n_visibles'] = np.sum(np.sum(np.isnan(y), axis=1) == 0, axis=0).astype(int)

    return d.convert_dtypes()


def merge(dfs, *, primary_key=['frame', 'label'], key_n_inliers='n_inliers', key_others=['X', 'Y', 'Z', 'reproj', 'outliers', 'inliers', 'n_outliers']):
    SUFFIX = 'tmp'
    KEYS = primary_key
    VALS = key_others + [key_n_inliers]
    VALS_y = [f'{v}_{SUFFIX}' for v in VALS]

    # merge points
    df = None
    for d in dfs:
        for c in KEYS + VALS:
            if not c in d:
                print(d)
                raise Exception(f'{d} does not have "{c}" column')

        d = d[KEYS+VALS]
        if df is None:
            df = d
            continue

        di = pd.merge(df, d, on=KEYS, how='outer', suffixes=['', f'_{SUFFIX}'])

        # import if the input has more inliers, or the point does not in the current list
        idx = (di[key_n_inliers] < di[f'{key_n_inliers}_{SUFFIX}']) | di[key_n_inliers].isna()
        #print(f'Found {len(di)} points in total (new {np.sum(idx)} points from {f})')
        di.loc[idx,VALS] = di.loc[idx,VALS_y].rename(columns={f'{v}_{SUFFIX}':v for v in VALS})
        df = di.loc[:,~di.columns.str.contains(f'_{SUFFIX}$', case=False)]

    df = df.sort_values(by=KEYS)
    return df


def triangulate_wrapper(x):
    df, P, c1, c2, reproj_th, primary_key, key_cid, key_x, key_y = x
    return triangulate(df, P, c1, c2, reproj_th=reproj_th, primary_key=primary_key, key_cid=key_cid, key_x=key_x, key_y=key_y)


def triangulate_inliers(df_undistorted, X, P, *, primary_key=['frame', 'label'], key_cid='camera', key_x='x', key_y='y'):
    inliers = X[primary_key + ['inliers']]
    df = pd.merge(df_undistorted, inliers, how='inner', on=primary_key)

    for c in sorted(df[key_cid].unique()):
        idx = np.array([(c in i and c == j) for i,j in zip(df['inliers'].to_list(), df[key_cid].to_list())])
        dfc = df.loc[idx]
        df = pd.merge(df, dfc[primary_key+[key_x, key_y]], how='left', on=primary_key, suffixes=['', f'_{c}'])
    df = df.drop(columns=[key_x, key_y, key_cid])
    df = df.drop_duplicates(primary_key)
    df = df.sort_values(primary_key)

    # do triangulation for each set of inliers
    df_all = []
    for inliers, d in df.groupby(by='inliers'):
        Pg = P[list(inliers)] # Nc x 3 x 4
        pt2d = []
        for c in inliers:
            p = d[[f'{key_x}_{c}', f'{key_y}_{c}']]
            pt2d.append(p.to_numpy())
        pt2d = np.array(pt2d) # Nc x Np x 2
        pt3d = pycalib.calib.triangulate_Npts(pt2d, Pg)
        d[['X', 'Y', 'Z']] = pt3d
        df_all.append(d)
    df = pd.concat(df_all).sort_values(primary_key)
    df = pd.merge(X, df[primary_key+['X','Y','Z']], how='inner', on=primary_key, suffixes=['', '_in'])
    return df


def triangulate_consensus(df, P, *, nproc=0, reproj_th=10, re_triangulate=True, show_pbar=True, primary_key=['frame', 'label'], key_cid='camera', key_x='x', key_y='y', distorted=False, camera_matrix=None, dist_coeffs=None):
    """
    Robust n-view triangulation

    Parameters
    ----------
    df: pandas.DataFrame
        list of 2D points with columns [primary_key, key_cid, key_x, key_y]
    P: numpy.array
        projection matrices (Nc x 3 x 4)
    nproc: int
        number of concurrent processes
    reproj_th: float
        reprojection error threshold
    re_triangulate: Bool
        do final triangulation using identified inliers
    show_pbar: Bool
        show progress bar
    primary_key: list of str
        primary key columns to identify points
    key_cid: str
        column name of camera id
    key_x: str
        column name of x
    key_y: str
        column name of y
    distorted: Bool
        do undistortion before triangulation (requires camera_matrix and dist_coeffs)
    camera_matrix: numpy.array
        camera matrix (Nc x 3 x 3)
    dist_coeffs: numpy.array
        distortion coefficients (Nc x {4,5,8})

    Returns
    -------
    X: pandas.DataFrame
        triangulated 3D points with columns [primary_key, key_cid, key_x, key_y, X, Y, Z]
    df: pandas.DataFrame
        undistorted 2D points with columns [primary_key, key_cid, key_x, key_y]
    """

    df = df[primary_key+[key_cid, key_x, key_y]].copy()

    cids = df[key_cid].unique()
    if distorted:
        df = df.copy()
        assert camera_matrix is not None and dist_coeffs is not None, 'camera_matrix and dist_coeffs must be provided'
        for c in cids:
            undistort_inplace(df, c, camera_matrix[c], dist_coeffs[c])

    # two-view triangulation in parallel
    p = multiprocessing.Pool(multiprocessing.cpu_count() if nproc <= 0 else nproc)
    args = [(df, P, c1, c2, reproj_th, primary_key, key_cid, key_x, key_y) for c1, c2 in itertools.combinations(cids, 2)]
    dfs = []
    for d in tqdm(p.imap_unordered(triangulate_wrapper, args), total=len(args), disable=not show_pbar):
        dfs.append(d)
    p.close()
    p.join()

    # merge results
    X = merge(dfs, primary_key=primary_key)

    # re-triangulation using all inliers
    if re_triangulate:
        X = triangulate_inliers(df, X, P, primary_key=primary_key, key_cid=key_cid, key_x=key_x, key_y=key_y)

    return X, df


def main_undistort():
    parser = argparse.ArgumentParser()
    parser.add_argument('calib', type=str, help='Calibration file (e.g., calib.json)')
    parser.add_argument('pt2d_distorted', type=str, help='2D points file (e.g., keypoint2d_allframe.csv)')
    parser.add_argument('output', type=str, help='Output filename (e.g., keypoint2d_allframe_undistorted.csv)')
    args = parser.parse_args()

    # Load calib
    K, D, R, T, _ = pycalib.util.load_calib(args.calib)
    print(f'Found {len(K)} cameras from {args.calib}')

    # Load 2D points
    df = pd.read_csv(args.pt2d_distorted).dropna()
    print(f'Found {len(df)} 2D points from {args.pt2d_distorted}')

    # Undistort for each camera
    for c in df['camera'].unique():
        undistort_inplace(df, c, K[c], D[c])

    # Save
    df.to_csv(args.output, index=False, float_format="%.16f")


def main_triangulate():
    parser = argparse.ArgumentParser()
    parser.add_argument('calib', type=str, help='Calibration file (e.g., calib.json)')
    parser.add_argument('pt2d_undistorted', type=str, help='2D points file (e.g., keypoint2d_allframe_undistorted.csv)')
    parser.add_argument('cam1', type=int, help='1st camera for triangulation')
    parser.add_argument('cam2', type=int, help='2nd camera for triangulation')
    parser.add_argument('output', type=str, help='Output csv file')
    parser.add_argument('-r', '--reproj_th', type=float, default=5, help='Max reprojection error to accept')
    args = parser.parse_args()

    # Load calib
    K, D, R, T, _ = pycalib.util.load_calib(args.calib)
    print(f'Found {len(K)} cameras from {args.calib}')
    P = []
    for k, r, t in zip(K, R, T):
        p = k @ np.hstack([r, t])
        P.append(p)
    P = np.array(P)

    # Load 2D points
    df = pd.read_csv(args.pt2d_undistorted).dropna()
    print(f'Found {len(df)} 2D points from {args.pt2d_undistorted}')
    if 'index' in df:
        df = df.drop(columns='index')

    d = triangulate(df, P, args.cam1, args.cam2, reproj_th=args.reproj_th)

    # Save
    d.to_csv(args.output, index=False)
    print(f'Saved to {args.output}')

def main_merge():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', nargs='+', required=True, type=str, help='CSV files with inliers and visible columns')
    parser.add_argument('-o', '--output', required=True, type=str, help='output CSV file')
    args = parser.parse_args()
    dfs = [pd.read_csv(f) for f in args.input]
    df = merge(dfs)
    df.to_csv(args.output, index=False, float_format="%.16f")
    print(f'Saved to {args.output}')

def main_triangulate_consensus():
    parser = argparse.ArgumentParser()
    parser.add_argument('calib', type=str, help='Calibration file (e.g., calib.json)')
    parser.add_argument('pt2d_undistorted', type=str, help='2D points file (e.g., keypoint2d_allframe_undistorted.csv)')
    parser.add_argument('output', type=str, help='Output csv file')
    parser.add_argument('-r', '--reproj_th', type=float, default=5, help='Max reprojection error to accept')
    args = parser.parse_args()

    # Load calib
    K, D, R, T, _ = pycalib.util.load_calib(args.calib)
    print(f'Found {len(K)} cameras from {args.calib}')
    P = []
    for k, r, t in zip(K, R, T):
        p = k @ np.hstack([r, t])
        P.append(p)
    P = np.array(P)

    # Load 2D points
    df = pd.read_csv(args.pt2d_undistorted).dropna()
    print(f'Found {len(df)} 2D points from {args.pt2d_undistorted}')
    if 'index' in df:
        df = df.drop(columns='index')

    d = triangulate_consensus(df, P, reproj_th=args.reproj_th)

    # Save
    d.to_csv(args.output, index=False)
    print(f'Saved to {args.output}')

if __name__ == '__main__':
    import argparse
    #main_undistort()
    #main_triangulate()
    #main_merge()
    main_triangulate_consensus()
